// Copyright (C) 2011  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_HASh_
#define DLIB_HASh_


#include "general_hash/hash.h"
#include "general_hash/random_hashing.h"
#include "general_hash/count_bits.h"


#endif // DLIB_HASh_


